from beartype import beartype
from typing import Dict, List, Tuple
import torch
from tqdm import tqdm

def compute_target_tail_indecies(triplets: List[Tuple[str, str, str]], tail_entitiy_ids: List[str])->List[int]:
    entity_id2index = {entity_id: index for index, entity_id in enumerate(tail_entitiy_ids)}
    target_tail_indecies: List[int] = [entity_id2index[triplets[i][2]] for i in range(len(triplets))]
    return target_tail_indecies

@beartype
def find_ranked_tails(batched_hr_embedding: torch.Tensor, tail_embeddings: torch.Tensor)->torch.Tensor:
    '''Find the ranked tails given the head-relation embedding and the tail embeddings'''
    # Compute cosine similarity between hr_embedding and tail_embeddings
        
    normalised_batched_hr_embeddings = batched_hr_embedding/batched_hr_embedding.norm(dim=1)[:, None]
    normalised_tail_embeddings = tail_embeddings/tail_embeddings.norm(dim=1)[:, None]
    similarities = torch.mm(normalised_batched_hr_embeddings, normalised_tail_embeddings.T)
    
    # Sort the similarities
    sorted_similarities, indecies = torch.sort(similarities, descending=True, dim=1)
    return indecies

@beartype
def compute_target_tail_ranks(hr_embeddings: torch.Tensor,
                              tail_embeddings: torch.Tensor,
                              target_tail_indecies: List[int])->List[int]:
    '''Predict the target tail ranks given the tsvs, target_tail_indecies, entity_id2text, relation_id2text, tail_embeddings, model_inputs, and model_name'''
    tail_ranks: List[int] = []
    number_of_triplets = hr_embeddings.shape[0]
    batch_size: int = 500
    batched_hr_embeddings = torch.split(hr_embeddings, batch_size)
    batched_target_tail_indecies = torch.split(torch.tensor(target_tail_indecies), batch_size)
    
    for batched_hr_embedding, batched_target_tail_indecie in tqdm(zip(batched_hr_embeddings, batched_target_tail_indecies), total=number_of_triplets//batch_size):
        batched_entity_ranks = find_ranked_tails(batched_hr_embedding, tail_embeddings)
        # Get Tail Ranks
        # The line above is really slow, optimize it in the future
        target_tail_ranks = [batched_entity_ranks[i].tolist().index(batched_target_tail_indecie[i]) for i in range(batched_entity_ranks.shape[0])]
        tail_ranks += target_tail_ranks
    return tail_ranks

@beartype
def compute_metrics(tail_ranks: List[int], ks: List[int] = [1, 5, 10])->Dict[str, float]:
    '''Compute the mean rank and the hit@10'''
    results: Dict[str, float] = {}
    results['Mean Rank'] = sum(tail_ranks)/len(tail_ranks)
    results['Mean Reciprical Rank'] = sum([1/rank for rank in tail_ranks])/len(tail_ranks)
    for k in ks:
        results[f'Hit@{k}'] = sum([1 for rank in tail_ranks if rank <= k])/len(tail_ranks)
    return results


from typing import Dict, List, Tuple
import torch
from beartype import beartype

from encoder_model import PromptEOL_Encoder
from prompt_templates import query2prompts, tail_entities2prompts
from post_processing import compute_target_tail_indecies, compute_target_tail_ranks, compute_metrics

@beartype
def save_encodings(embeddings:List[torch.Tensor], embeddings_save_path:str):
    '''
    Save the embeddings to a file.
    
    Args:
    embeddings: List of torch.Tensor
    embeddings_save_path: str
    '''
    concatenated_embeddings = torch.concat(embeddings, dim=0)
    concatenated_embeddings = concatenated_embeddings.numpy()
    print('Shape of concatenated embeddings')
    print(concatenated_embeddings.shape)
    torch.save(concatenated_embeddings, embeddings_save_path, pickle_protocol=4)
    
def knowledge_probe(triplets: List[Tuple[str, str, str]], entity_id2text: Dict[str, str],
                    relation_id2text:Dict[str, str], entity_id2definition:Dict[str, str], fewshot_prompt:str, 
                    model_name: str, cuda: bool )->Dict[str, float]:
    encoder = PromptEOL_Encoder(model_name, cuda=cuda)    
    query_prompts = query2prompts(triplets,
                entity_id2text,
                relation_id2text, 
                fewshot_prompt = fewshot_prompt,
                entity_id2definition = entity_id2definition)
    query_encodings = encoder(query_prompts)
    
    tail_prompts = tail_entities2prompts(list(entity_id2text.values()), list(entity_id2definition.values())) 
    tail_encodings = encoder(tail_prompts)
    
    target_tail_indecies:List[int] = compute_target_tail_indecies(triplets, list(entity_id2text.keys()))
    target_tail_ranks: List[int] = compute_target_tail_ranks(query_encodings, tail_encodings, target_tail_indecies)
    results: dict[str, float] = compute_metrics(target_tail_ranks)
    return results


'''
Contains code for extracting the last hidden vector from a transformer model.
'''

from typing import List, TypedDict
import torch
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
from transformers.modeling_utils import PreTrainedModel
from transformers import AutoModelForCausalLM, AutoTokenizer
from tqdm import tqdm
from beartype import beartype

class LLMInputs(TypedDict):
    input_ids: torch.Tensor
    attention_mask: torch.Tensor
    
    
class PromptEOL_Encoder:
    '''
    A class to encode prompts using a transformer model.
    '''
    @beartype
    def __init__(self, model_name:str, cuda:bool=False):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        if cuda==False:
            self.model = AutoModelForCausalLM.from_pretrained(model_name)
            self.device: str = 'cpu'
            print('Using CPU')
        else:
            number_of_devices = torch.cuda.device_count()
            print(f'Number of devices: {number_of_devices}')
            self.model = AutoModelForCausalLM.from_pretrained(model_name, device_map="auto",  # Distributes layers across GPUs
            offload_folder=None )
            self.device = torch.cuda.current_device()
            
        self.model.eval()
        
    def __call__(self, prompts: List[str])->torch.Tensor:
        llm_inputs: List[LLMInputs] = PromptEOL_Encoder.text2llm_inputs(self.tokenizer, prompts)
        encodings: torch.Tensor = PromptEOL_Encoder.encode(llm_inputs, self.model, self.device)
        
        return encodings
        
    @beartype
    @staticmethod
    def text2llm_inputs(tokenizer: PreTrainedTokenizerBase, prompts: List[str])->List[LLMInputs]:
        llm_inputs = []
        for prompt in tqdm(prompts, total=len(prompts)):
            llm_inputs.append(tokenizer(prompt, return_tensors="pt"))
        return llm_inputs
        
    @beartype
    @staticmethod
    def encode(llm_inputs: List[LLMInputs], model: PreTrainedModel, device: str)->torch.Tensor:
        last_vectors: List[torch.Tensor] = []
        with torch.no_grad():
            for llm_input in tqdm(llm_inputs, total=len(llm_inputs)):
                llm_input = {k: v.to(device) for k, v in llm_input.items()}
                outputs = model(**llm_input, output_hidden_states=True)

                # Get the last hidden state vector
                last_hidden_state = outputs.hidden_states[-1]  # Last layer's hidden states

                # Extract the vector corresponding to the last token
                last_vector = last_hidden_state[:, -1, :]
                if device!='cpu':
                    last_vector = last_vector.cpu()
                last_vectors.append(last_vector)
        last_vectors = torch.concat(last_vectors, dim=0)
        return last_vectors


'''
Contains code for creating prompts
'''

from typing import Dict, List, Optional, Tuple
from beartype import beartype
from tqdm import tqdm

@beartype
def query2prompts(triplets: List[Tuple[str, str, str]],
                    entity_id2text: Dict[str, str], 
                    relation_id2text: Dict[str, str], 
                    fewshot_prompt: Optional[str] = False,
                    entity_id2definition: Optional[Dict[str, str]] = None
                              )->List[str]:
    head_entity_ids, relation_ids, tail_entity_ids = zip(*triplets)
    prompts: List[str] = []
    if entity_id2definition is not None:
        print('Using definitions')
    for i in range(len(triplets)):
        head_entity = entity_id2text[head_entity_ids[i]]
        relation = relation_id2text[relation_ids[i]]
        text = f'''\n({head_entity}, {relation}, '''
        if entity_id2definition is not None:
            definition = entity_id2definition[head_entity_ids[i]]
            text = f'''{head_entity} - {definition}\n({head_entity}, {relation}, '''
        if fewshot_prompt:
            text = f'''{fewshot_prompt}\n{text}'''
        prompts.append(text)
    return prompts

def tail_entities2prompts(entities: List[str], definitions: Optional[List[str]])->List[str]:
    prompts: List[str] = []
    for i, entity in tqdm(enumerate(entities), total=len(entities)):
        
        prompt_sent = f"This sentence: \"{entity}\" means in one word: \""
        if definitions is not None:
            prompt_sent = f"{entity} - {definitions[i]}"+'\nThis sentence: "{word}" means in one word: "{one word}"'+f"\nThis sentence: \"{entity}\" means in one word: \""
        prompts.append(prompt_sent)
    return prompts